from tabulate import tabulate

from ...agent import Agent
from ...exceptions import BackendClientError
from ..pretty import print_fail
from . import admin


@admin.register_command
def agents(args):
    '''
    List and manage agents.
    '''
    fields = [
        ('ID', 'id'),
        ('Status', 'status'),
        ('First Contact', 'first_contact'),
        ('Mem.Slots', 'mem_slots'),
        ('CPU Slots', 'cpu_slots'),
        ('GPU Slots', 'gpu_slots'),
    ]
    try:
        items = Agent.list(args.status,
                           fields=(item[1] for item in fields))
    except BackendClientError as e:
        print_fail(str(e))
        return
    if len(items) == 0:
        print('There are no matching agents.')
        return
    print(tabulate((item.values() for item in items),
                   headers=(item[0] for item in fields)))


agents.add_argument('-s', '--status', type=str, default='ALIVE',
                    help='Filter agents by the given status.')



from . import register_command
from ..config import get_config


@register_command
def config(args):
    '''
    Shows the current configuration.
    '''
    config = get_config()
    print('API endpoint: "{0}"'.format(config.endpoint))
    print('API version: "{0}"'.format(config.version))
    print('Access key: "{0}"'.format(config.access_key))
    masked_skey = config.secret_key[:6] + ('*' * 24) + config.secret_key[-10:]
    print('Secret key: "{0}"'.format(masked_skey))
    print('Signature hash type: "{0}"'.format(config.hash_type))



from pathlib import Path
import re
from typing import Sequence, Union

import aiohttp

from .base import BaseFunction, SyncFunctionMixin
from .request import Request

__all__ = (
    'BaseVFolder',
    'VFolder',
)

_rx_slug = re.compile(r'^[a-zA-Z0-9]([a-zA-Z0-9._-]*[a-zA-Z0-9])?$')


class BaseVFolder(BaseFunction):
    @classmethod
    def _create(cls, name: str):
        assert _rx_slug.search(name) is not None
        resp = yield Request('POST', '/folders/', {
            'name': name,
        })
        return resp.json()

    @classmethod
    def _list(cls):
        resp = yield Request('GET', '/folders/')
        return resp.json()

    @classmethod
    def _get(cls, name: str):
        return cls(name)

    def _info(self):
        resp = yield Request('GET', '/folders/{0}'.format(self.name))
        return resp.json()

    def _delete(self):
        resp = yield Request('DELETE', '/folders/{0}'.format(self.name))
        if resp.status == 200:
            return resp.json()

    def _upload(self, files: Sequence[Union[str, Path]],
               basedir: Union[str, Path]=None):
        fields = []
        base_path = (Path.cwd() if basedir is None
                     else Path(basedir).resolve())
        for file in files:
            file_path = Path(file).resolve()
            try:
                fields.append(aiohttp.web.FileField(
                    'src',
                    str(file_path.relative_to(base_path)),
                    open(str(file_path), 'rb'),
                    'application/octet-stream',
                    None
                ))
            except ValueError:
                msg = 'File "{0}" is outside of the base directory "{1}".' \
                      .format(file_path, base_path)
                raise ValueError(msg) from None
        rqst = Request('POST', '/folders/{}/upload'.format(self.name))
        rqst.content = fields
        resp = yield rqst
        return resp

    def __init__(self, name: str):
        assert _rx_slug.search(name) is not None
        self.name = name
        self.delete = self._call_base_method(self._delete)
        self.info = self._call_base_method(self._info)
        self.upload = self._call_base_method(self._upload)

    def __init_subclass__(cls):
        cls.create = cls._call_base_clsmethod(cls._create)
        cls.list = cls._call_base_clsmethod(cls._list)
        cls.get = cls._call_base_clsmethod(cls._get)


class VFolder(SyncFunctionMixin, BaseVFolder):
    pass



from typing import Optional, Iterable

from .base import BaseFunction, SyncFunctionMixin
from .request import Request

__all__ = (
    'BaseAgent',
    'Agent',
)


class BaseAgent(BaseFunction):

    @classmethod
    def _list(cls,
              status: str='ALIVE',
              fields: Optional[Iterable[str]]=None):
        if fields is None:
            fields = (
                'id',
                'addr',
                'status',
                'first_contact',
                'mem_slots',
                'cpu_slots',
                'gpu_slots',
            )
        q = 'query($status: String) {' \
            '  agents(status: $status) {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(fields))
        vars = {
            'status': status,
        }
        resp = yield Request('POST', '/admin/graphql', {
            'query': q,
            'variables': vars,
        })
        data = resp.json()
        return data['agents']

    def __init_subclass__(cls):
        cls.list = cls._call_base_clsmethod(cls._list)


class Agent(SyncFunctionMixin, BaseAgent):
    pass



import enum
import functools
import sys
import textwrap

__all__ = (
    'PrintStatus', 'print_pretty', 'print_info', 'print_wait',
    'print_done', 'print_fail',
)


class PrintStatus(enum.Enum):
    NONE = 0
    WAITING = 1
    DONE = 2
    FAILED = 3


def print_pretty(msg, *, status=PrintStatus.NONE, file=None):
    if file is None:
        file = sys.stderr
    if status == PrintStatus.NONE:
        indicator = '\x1b[96m\u2219' if file.isatty() else '\u2219'
    elif status == PrintStatus.WAITING:
        assert '\n' not in msg, 'Waiting message must be a single line.'
        indicator = '\x1b[93m\u22EF' if file.isatty() else '\u22EF'
    elif status == PrintStatus.DONE:
        indicator = '\x1b[92m\u2714' if file.isatty() else '\u2714'
    elif status == PrintStatus.FAILED:
        indicator = '\x1b[91m\u2718' if file.isatty() else '\u2718'
    else:
        raise ValueError
    if file.isatty():
        print('\x1b[2K', end='', file=file)
    text = textwrap.indent(msg, '  ')
    text = indicator + text[1:]
    if file.isatty():
        text += '\x1b[0m'
    print('{0}\r'.format(text), end='', file=file)
    file.flush()
    if status != PrintStatus.WAITING:
        print('', file=file)


print_info = functools.partial(print_pretty, status=PrintStatus.NONE)
print_wait = functools.partial(print_pretty, status=PrintStatus.WAITING)
print_done = functools.partial(print_pretty, status=PrintStatus.DONE)
print_fail = functools.partial(print_pretty, status=PrintStatus.FAILED)



from argparse import Namespace

from . import register_command
from .admin.sessions import sessions


@register_command
def ps(args):
    '''
    Lists the current running compute sessions for the current keypair.
    This is an alias of the "admin sessions --status=RUNNING" command.
    '''
    inner_args = Namespace()
    inner_args.status = 'RUNNING'
    inner_args.access_key = None
    sessions(inner_args)



from . import register_command
from .pretty import print_wait, print_done, print_fail
from ..exceptions import BackendError
from ..kernel import Kernel


@register_command
def logs(args):
    '''
    Shows the output logs of a running container.
    '''
    try:
        print_wait('Retrieving container logs...')
        kernel = Kernel(args.sess_id_or_alias)
        result = kernel.get_logs().get('result')
        logs = result.get('logs') if 'logs' in result else ''
        print(logs)
        print_done('End of logs.')
    except BackendError as e:
        print_fail(str(e))
        return


logs.add_argument('sess_id_or_alias', metavar='NAME',
                  help='The session ID or its alias '
                       'given when creating the session.')



from .exceptions import *  # noqa
from .admin import *  # noqa
from .kernel import *  # noqa
from .vfolder import *  # noqa
from .request import *  # noqa

__all__ = [
    exceptions.__all__,  # noqa
    admin.__all__,  # noqa
    kernel.__all__,  # noqa
    vfolder.__all__,  # noqa
    request.__all__,  # noqa
]

__version__ = '1.2.0'



from tabulate import tabulate

from ...admin import Admin
from ...exceptions import BackendClientError
from ..pretty import print_fail
from . import admin


@admin.register_command
def sessions(args):
    '''
    List and manage compute sessions.
    '''
    fields = [
        ('Session ID', 'sess_id'),
        ('Lang/runtime', 'lang'),
        ('Created At', 'created_at',),
        ('Termianted At', 'terminated_at'),
        ('Status', 'status'),
        ('Memory Slot', 'mem_slot'),
        ('CPU Slot', 'cpu_slot'),
        ('GPU Slot', 'gpu_slot'),
    ]
    if args.access_key is None:
        q = 'query($status:String) { compute_sessions(status:$status) { $fields } }'
    else:
        q = 'query($ak:String, $status:String) {' \
            '  compute_sessions(access_key:$ak, status:$status) { $fields }' \
            '}'
    q = q.replace('$fields', ' '.join(item[1] for item in fields))
    v = {
        'status': args.status if args.status != 'ALL' else None,
        'ak': args.access_key,
    }
    try:
        resp = Admin.query(q, v)
    except BackendClientError as e:
        print_fail(str(e))
        return
    if len(resp['compute_sessions']) == 0:
        print('There are no compute sessions currently running.')
        return
    print(tabulate((item.values() for item in resp['compute_sessions']),
                   headers=(item[0] for item in fields)))


sessions.add_argument('--status', type=str, default='RUNNING',
                      choices={'PREPARING', 'BUILDING', 'RUNNING',
                               'RESTARTING', 'RESIZING', 'SUSPENDED',
                               'TERMINATING', 'TERMINATED', 'ERROR', 'ALL'},
                      help='Filter by the given status (default: RUNNING)')
sessions.add_argument('--access-key', type=str, default=None,
                      help='Get sessions for a specific access key '
                           '(only works if you are a super-admin)')



from typing import Optional, Iterable

from .base import BaseFunction, SyncFunctionMixin
from .request import Request

__all__ = (
    'BaseKeyPair',
    'KeyPair',
)


class BaseKeyPair(BaseFunction):

    @classmethod
    def _create(cls, user_id: int,
                is_active: bool=True,
                is_admin: bool=False,
                resource_policy: Optional[str]=None,
                rate_limit: Optional[int]=None,
                concurrency_limit: Optional[int]=None,
                fields: Optional[Iterable[str]]=None):
        if fields is None:
            fields = ('access_key', 'secret_key')
        q = 'mutation($user_id: Int!, $input: KeyPairInput!) {' \
            '  create_keypair(user_id: $user_id, props: $input) {' \
            '    ok msg keypair { $fields }' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(fields))
        vars = {
            'user_id': user_id,
            'input': {
                'is_active': is_active,
                'is_admin': is_admin,
                'resource_policy': resource_policy,
                'rate_limit': rate_limit,
                'concurrency_limit': concurrency_limit,
            },
        }
        resp = yield Request('POST', '/admin/graphql', {
            'query': q,
            'variables': vars,
        })
        data = resp.json()
        return data['create_keypair']

    @classmethod
    def _list(cls, user_id: int,
              is_active: Optional[bool]=None,
              fields: Optional[Iterable[str]]=None):
        if fields is None:
            fields = (
                'access_key', 'secret_key',
                'is_active', 'is_admin',
            )
        q = 'query($user_id: Int!, $is_active: Boolean) {' \
            '  keypairs(user_id: $user_id, is_active: $is_active) {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(fields))
        vars = {
            'user_id': user_id,
            'is_active': is_active,
        }
        resp = yield Request('POST', '/admin/graphql', {
            'query': q,
            'variables': vars,
        })
        data = resp.json()
        return data['keypairs']

    @classmethod
    def activate(cls, access_key: str):
        raise NotImplementedError

    @classmethod
    def deactivate(cls, access_key: str):
        raise NotImplementedError

    def __init_subclass__(cls):
        cls.create = cls._call_base_clsmethod(cls._create)
        cls.list = cls._call_base_clsmethod(cls._list)


class KeyPair(SyncFunctionMixin, BaseKeyPair):
    pass



import os
from typing import Optional

__all__ = [
    'APIConfig',
]

_config = None


def get_env(name, default=None):
    v = os.environ.get('BACKEND_' + name)
    if v is None:
        v = os.environ.get('SORNA_' + name)
    if v is None:
        if default is None:
            raise KeyError(name)
        v = default
    return v


class APIConfig:

    DEFAULTS = {
        'endpoint': 'https://api.backend.ai',
        'user_agent': 'Backend.AI Client Library (Python/v1.0)',
        'version': 'v2.20170315',
    }

    def __init__(self, endpoint: Optional[str]=None,
                 version: str='v2.20170315',
                 user_agent: Optional[str]=None,
                 access_key: Optional[str]=None,
                 secret_key: Optional[str]=None,
                 hash_type: Optional[str]=None) -> None:
        self._endpoint = \
            endpoint if endpoint else get_env('ENDPOINT', self.DEFAULTS['endpoint'])
        self._version = \
            version if version else self.DEFAULTS['version']
        self._user_agent = \
            user_agent if user_agent else self.DEFAULTS['user_agent']
        self._access_key = \
            access_key if access_key else get_env('ACCESS_KEY')
        self._secret_key = \
            secret_key if secret_key else get_env('SECRET_KEY')
        self._hash_type = \
            hash_type.lower() if hash_type else 'sha256'

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def user_agent(self):
        return self._user_agent

    @property
    def access_key(self):
        return self._access_key

    @property
    def secret_key(self):
        return self._secret_key

    @property
    def version(self):
        return self._version

    @property
    def hash_type(self):
        return self._hash_type


def get_config():
    global _config
    if _config is None:
        _config = APIConfig()
    return _config


def set_config(conf: APIConfig):
    global _config
    _config = conf



import http.server
import re
import traceback

from . import register_command
from .pretty import print_info, print_fail
from ..request import Request


class APIProxyHandler(http.server.BaseHTTPRequestHandler):

    def do_proxy(self):
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)
        path = re.sub(r'^/?v(\d+)/', '/', self.path)
        rqst = Request(self.command, path, body)
        resp = rqst.send()
        self.send_response(resp.status, resp.reason)
        self.send_header('Content-Type', resp.content_type)
        self.send_header('Content-Length', resp.content_length)
        self.end_headers()
        self.wfile.write(resp.content)

    def do_HEAD(self):
        self.do_proxy()

    def do_GET(self):
        self.do_proxy()

    def do_POST(self):
        self.do_proxy()

    def do_PUT(self):
        self.do_proxy()

    def do_DELETE(self):
        self.do_proxy()

    def do_PATCH(self):
        self.do_proxy()

    def do_OPTIONS(self):
        self.do_proxy()


@register_command
def proxy(args):
    '''
    Run a non-encyprted non-authorized API proxy server.
    Use this only for development and testing!
    '''
    print_info('Starting an insecure API proxy at http://{0}:{1}'
               .format(args.bind, args.port))
    try:
        addr = (args.bind, args.port)
        httpd = http.server.HTTPServer(addr, APIProxyHandler)
        httpd.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        print()
        print_info('Terminated.')
    except:
        print_fail('Unexpected error!')
        traceback.print_exc()


proxy.add_argument('--bind', type=str, default='localhost',
                   help='The IP/host address to bind this proxy.')
proxy.add_argument('-p', '--port', type=int, default=8084,
                   help='The TCP port to accept non-encrypted non-authorized '
                        'API requests. (default: 8084)')



from tabulate import tabulate

from ...keypair import KeyPair
from ...exceptions import BackendClientError
from ..pretty import print_fail
from . import admin


@admin.register_command
def keypairs(args):
    '''
    List and manage keypairs.
    '''
    fields = [
        ('Access Key', 'access_key'),
        ('Secret Key', 'secret_key'),
        ('Active?', 'is_active'),
        ('Admin?', 'is_admin'),
        ('Created At', 'created_at'),
        ('Last Used', 'last_used'),
        ('Res.Policy', 'resource_policy'),
        ('Rate Limit', 'rate_limit'),
        ('Concur.Limit', 'concurrency_limit'),
    ]
    try:
        items = KeyPair.list(args.user_id, args.is_active,
                             fields=(item[1] for item in fields))
    except BackendClientError as e:
        print_fail(str(e))
        return
    if len(items) == 0:
        print('There are no matching keypairs associated '
              'with the user ID {0}'.format(args.user_id))
        return
    print(tabulate((item.values() for item in items),
                   headers=(item[0] for item in fields)))


keypairs.add_argument('-u', '--user-id', type=int, default=0,
                      help='Show keypairs of this given user. '
                           '[default: 0]')
keypairs.add_argument('--is-active', type=bool, default=None,
                      help='Filter keypairs by activation.')


@keypairs.register_command
def add(args):
    '''
    Add a new keypair.
    '''
    if args.user_id is None:
        print('You must set the user ID (-u/--user-id).')
        return
    try:
        data = KeyPair.create(args.user_id)
    except BackendClientError as e:
        print_fail(str(e))
        return
    if not data['ok']:
        print_fail('KeyPair creation has failed: {0}'
                   .format(data['msg']))
        return
    item = data['keypair']
    print('Access Key: {0}'.format(item['access_key']))
    print('Secret Key: {0}'.format(item['secret_key']))


add.add_argument('-u', '--user-id', type=int, default=None,
                 help='Create a keypair for this user.')



from pathlib import Path

from . import register_command
from ..vfolder import VFolder


@register_command
def vfolder(args):
    '''Provides virtual folder operations.'''
    print('Run with -h/--help for usage.')


@vfolder.register_command
def list(args):
    '''List virtual folders that belongs to the current user.'''
    data = VFolder.list()
    print(data)


@vfolder.register_command
def create(args):
    '''Create a new virtual folder.'''
    result = VFolder.create(args.name)
    print(result)


create.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def delete(args):
    '''Delete the given virtual folder. This operation is irreversible!'''
    VFolder(args.name).delete()


delete.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def info(args):
    '''Show the information of the given virtual folder.'''
    result = VFolder(args.name).info()
    print(result)


info.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def upload(args):
    '''Upload a file to the virtual folder.'''
    result = VFolder(args.name).upload(args.filenames)
    print(result.status)


upload.add_argument('name', type=str, help='The name of a virtual folder.')
upload.add_argument('filenames', type=Path, nargs='+',
                    help='Paths of the files to be uploaded.')



import getpass
from pathlib import Path
import sys
import traceback

from humanize import naturalsize
from tabulate import tabulate

from . import register_command
from ..compat import token_hex
from ..exceptions import BackendError
from .pretty import print_info, print_wait, print_done, print_fail
from ..kernel import Kernel


def exec_loop(kernel, code, mode, opts=None,
              vprint_wait=print_wait, vprint_done=print_done):
    opts = opts if opts else {}
    run_id = None  # use server-assigned run ID
    while True:
        result = kernel.execute(run_id, code, mode=mode, opts=opts)
        run_id = result['runId']
        opts.clear()  # used only once
        for rec in result['console']:
            if rec[0] == 'stdout':
                print(rec[1], end='', file=sys.stdout)
            elif rec[0] == 'stderr':
                print(rec[1], end='', file=sys.stderr)
            else:
                print('----- output record (type: {0}) -----'.format(rec[0]))
                print(rec[1])
                print('----- end of record -----')
        sys.stdout.flush()
        if result['status'] == 'build-finished':
            exitCode = result.get('exitCode')
            vprint_done('Build finished. (exit code = {0})'.format(exitCode))
            mode = 'continue'
            code = ''
        elif result['status'] == 'finished':
            exitCode = result.get('exitCode')
            vprint_done('Finished. (exit code = {0})'.format(exitCode))
            break
        elif result['status'] == 'waiting-input':
            mode = 'input'
            if result['options'].get('is_password', False):
                code = getpass.getpass()
            else:
                code = input()
        elif result['status'] == 'continued':
            mode = 'continue'
            code = ''


def _noop(*args, **kwargs):
    pass


def _format_stats(stats):
    formatted = []
    for k, v in stats.items():
        if k.endswith('_size') or k.endswith('_bytes'):
            v = naturalsize(v, binary=True)
        elif k == 'cpu_used':
            k += '_msec'
            v = '{0:,}'.format(int(v))
        else:
            v = '{0:,}'.format(int(v))
        formatted.append((k, v))
    return tabulate(formatted)


@register_command
def run(args):
    '''
    Run the given code snippet or files in a session.
    Depending on the session ID you give (default is random),
    it may reuse an existing session or create a new one.
    '''
    if args.quiet:
        vprint_info = vprint_wait = vprint_done = _noop
    else:
        vprint_info = print_info
        vprint_wait = print_wait
        vprint_done = print_done
    if not args.client_token:
        args.client_token = token_hex(16)
        vprint_wait('Creating a temporary kernel...')
    else:
        vprint_info('Client session token: {0}'.format(args.client_token))
        vprint_wait('Connecting to the kernel...')
    if args.env is not None:
        envs = {k: v for k, v in map(lambda s: s.split('=', 1), args.env)}
    else:
        envs = {}
    try:
        kernel = Kernel.get_or_create(
            args.lang, args.client_token,
            mounts=args.mount,
            envs=envs)
    except BackendError as e:
        print_fail(str(e))
        return
    if kernel.created:
        vprint_done('Session {0} is ready.'.format(kernel.kernel_id))
    else:
        vprint_done('Reusing session {0}...'.format(kernel.kernel_id))

    try:
        if args.files:
            if args.code:
                print('You can run only either source files or command-line '
                      'code snippet.', file=sys.stderr)
                return
            vprint_wait('Uploading source files...')
            ret = kernel.upload(args.files, basedir=args.basedir)
            if ret.status // 100 != 2:
                print_fail('Uploading source files failed!')
                print('{0}: {1}\n{2}'.format(
                    ret.status, ret.reason, ret.text()))
                return
            vprint_done('Uploading done.')
            build_cmd = args.build if args.build else '*'
            exec_cmd = args.exec if args.exec else '*'
            exec_loop(kernel, '', 'batch', opts={
                'build': build_cmd,
                'exec': exec_cmd,
            }, vprint_wait=vprint_wait, vprint_done=vprint_done)
        else:
            if not args.code:
                print('You should provide the command-line code snippet using '
                      '"-c" option if run without files.', file=sys.stderr)
                return
            exec_loop(kernel, args.code, 'query',
                      vprint_wait=vprint_wait, vprint_done=vprint_done)
    except BackendError as e:
        print_fail(str(e))
    except Exception:
        print_fail('Execution failed!')
        traceback.print_exc()
    finally:
        if args.rm:
            vprint_wait('Cleaning up the session...')
            ret = kernel.destroy()
            vprint_done('Cleaned up the session.')
            if args.stats:
                stats = ret.get('stats', None) if ret else None
                if stats:
                    print(_format_stats(stats))
                else:
                    print('Statistics is not available.')


run.add_argument('lang',
                 help='The runtime or programming language name')
run.add_argument('files', nargs='*', type=Path,
                 help='The code file(s). Can be added multiple times')
run.add_argument('-t', '--client-token', metavar='SESSID',
                 help='Specify a human-readable session ID or name '
                      '[default: a random hex-string]')
run.add_argument('-c', '--code', metavar='CODE',
                 help='The code snippet as a single string')
run.add_argument('--build', metavar='CMD',
                 help='Custom shell command for building the given files')
run.add_argument('--exec', metavar='CMD',
                 help='Custom shell command for executing the given files')
run.add_argument('--basedir', metavar='PATH', type=Path, default=None,
                 help='Base directory path of uploaded files.  '
                      'All uploaded files must reside inside this directory.')
run.add_argument('--rm', action='store_true', default=False,
                 help='Terminate the session immediately after running '
                      'the given code or files [default: False]')
run.add_argument('-e', '--env', metavar='KEY=VAL', type=str, action='append',
                 help='Environment variable '
                      '(may appear multiple times)')
run.add_argument('-m', '--mount', type=str, action='append',
                 help='User-owned virtual folder names to mount')
run.add_argument('-s', '--stats', action='store_true', default=False,
                 help='Show resource usage statistics after termination '
                      '(only works if "--rm" is given)')
run.add_argument('-q', '--quiet', action='store_true', default=False,
                 help='Hide execution details but show only the kernel'
                      'outputs')


@register_command
def terminate(args):
    '''
    Terminate the given session.
    '''
    print_wait('Terminating the session...')
    try:
        kernel = Kernel(args.sess_id_or_alias)
        ret = kernel.destroy()
    except BackendError as e:
        print_fail(str(e))
        return
    else:
        print_done('Done.')
        if args.stats:
            stats = ret.get('stats', None) if ret else None
            if stats:
                print(_format_stats(stats))
            else:
                print('Statistics is not available.')


terminate.add_argument('sess_id_or_alias', metavar='NAME',
                       help='The session ID or its alias '
                            'given when creating the session.')
terminate.add_argument('-s', '--stats', action='store_true', default=False,
                       help='Show resource usage statistics after termination')



from abc import abstractmethod
import functools
import inspect

from .compat import Py36Object
from .exceptions import BackendAPIError

__all__ = (
    'BaseFunction',
    'SyncFunctionMixin',
    'AsyncFunctionMixin',
)


class BaseFunction(Py36Object):

    '''
    Implements the request creation and response handling logic,
    while delegating the process of request sending to the subclasses
    via the generator protocol.
    '''

    @abstractmethod
    def _call_base_method(self, meth):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _call_base_clsmethod(cls, meth):
        raise NotImplementedError

    @staticmethod
    def _handle_response(resp, meth_gen):
        if resp.status // 100 != 2:
            raise BackendAPIError(resp.status, resp.reason, resp.text())
        try:
            meth_gen.send(resp)
        except StopIteration as e:
            return e.value
        else:
            raise RuntimeError('Invalid state')


class SyncFunctionMixin:
    '''
    Synchronous request sender using requests.
    '''

    @staticmethod
    def _make_request(gen):
        rqst = next(gen)
        resp = rqst.send()
        return resp

    @classmethod
    def _call_base_clsmethod(cls, meth):
        assert inspect.ismethod(meth)

        @classmethod
        @functools.wraps(meth)
        def _caller(cls, *args, **kwargs):
            gen = meth(*args, **kwargs)
            resp = cls._make_request(gen)
            return cls._handle_response(resp, gen)

        return _caller

    def _call_base_method(self, meth):
        assert inspect.ismethod(meth)

        @functools.wraps(meth)
        def _caller(*args, **kwargs):
            gen = meth(*args, **kwargs)
            resp = self._make_request(gen)
            return self._handle_response(resp, gen)

        return _caller


class AsyncFunctionMixin:
    '''
    Asynchronous request sender using aiohttp.
    '''

    @staticmethod
    async def _make_request(gen):
        rqst = next(gen)
        resp = await rqst.asend()
        return resp

    @classmethod
    def _call_base_clsmethod(cls, meth):
        assert inspect.ismethod(meth)

        @classmethod
        @functools.wraps(meth)
        async def _caller(cls, *args, **kwargs):
            gen = meth(*args, **kwargs)
            resp = await cls._make_request(gen)
            return cls._handle_response(resp, gen)

        return _caller

    def _call_base_method(self, meth):
        assert inspect.ismethod(meth)

        @functools.wraps(meth)
        async def _caller(*args, **kwargs):
            gen = meth(*args, **kwargs)
            resp = await self._make_request(gen)
            return self._handle_response(resp, gen)

        return _caller



from tabulate import tabulate

from ...admin import Admin
from ...exceptions import BackendClientError
from ..pretty import print_fail
from . import admin


@admin.register_command
def vfolders(args):
    '''
    List and manage virtual folders.
    '''
    fields = [
        ('Name', 'name'),
        ('Created At', 'created_at'),
        ('Last Used', 'last_used'),
        ('Max Files', 'max_files'),
        ('Max Size', 'max_size'),
    ]
    if args.access_key is None:
        q = 'query { vfolders { $fields } }'
    else:
        q = 'query($ak:String) { vfolders(access_key:$ak) { $fields } }'
    q = q.replace('$fields', ' '.join(item[1] for item in fields))
    try:
        resp = Admin.query(q)
    except BackendClientError as e:
        print_fail(str(e))
        return
    print(tabulate((item.values() for item in resp['vfolders']),
                   headers=(item[0] for item in fields)))


vfolders.add_argument('--access-key', type=str, default=None,
                      help='Get vfolders for a specific access key '
                           '(only works if you are a super-admin)')



import json

import aiohttp
import aiohttp.web

from ..base import AsyncFunctionMixin
from ..exceptions import BackendClientError
from ..request import Request
from ..kernel import BaseKernel

__all__ = (
    'Kernel',
    'AsyncKernel',
    'StreamPty',
)


class Kernel(AsyncFunctionMixin, BaseKernel):

    # only supported in AsyncKernel
    async def stream_pty(self):
        request = Request('GET', '/stream/kernel/{}/pty'.format(self.kernel_id))
        try:
            sess, ws = await request.connect_websocket()
        except aiohttp.ClientResponseError as e:
            raise BackendClientError(e.code, e.message)
        return StreamPty(self.kernel_id, sess, ws)


class StreamPty:

    '''
    A very thin wrapper of aiohttp.WebSocketResponse object.
    It keeps the reference to the mother aiohttp.ClientSession object while the
    connection is alive.
    '''

    def __init__(self, kernel_id, sess, ws):
        self.kernel_id = kernel_id
        self.sess = sess  # we should keep reference while connection is active.
        self.ws = ws

    @property
    def closed(self):
        return self.ws.closed

    def __aiter__(self):
        return self.ws.__aiter__()

    async def __anext__(self):
        msg = await self.ws.__anext__()
        return msg

    def exception(self):
        return self.ws.exception()

    async def send_str(self, raw_str):
        await self.ws.send_str(raw_str)

    async def resize(self, rows, cols):
        await self.ws.send_str(json.dumps({
            'type': 'resize',
            'rows': rows,
            'cols': cols,
        }))

    async def restart(self):
        await self.ws.send_str(json.dumps({
            'type': 'restart',
        }))

    async def close(self):
        await self.ws.close()
        await self.sess.close()


# legacy alias
AsyncKernel = Kernel
